#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This module provides unit tests for `ptutils.debug`."""

import re


def test_stacktrace():
    """ Test that the debug stacktrace function produces a proper stacktrace. """
    from ptutils.debug import stacktrace
    try:
        raise Exception("intentional exception")
    except:  # noqa: E722
        st = stacktrace()

    assert isinstance(st, str)

    st_lines = st.split('\n')
    assert len(st_lines) == 5
    assert st_lines[0]  == "Traceback (most recent call last):"
    assert re.match(r'^ +File "[^"]+", line \d+, in test_stacktrace$', st_lines[-4])
    assert re.match(r'^ +raise Exception\("intentional exception\"\)', st_lines[-3])
    assert st_lines[-2] == "Exception: intentional exception"
    assert st_lines[-1] == ""



#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides unit tests for `ptutils.undefined`.
"""

from ptutils.undefined import is_defined, is_undefined, default_if_undefined, UNDEFINED, Undefined


def test_is_undefined():
    assert is_undefined(UNDEFINED)
    assert is_undefined(Undefined)
    assert is_undefined(Undefined())
    assert not is_undefined(None)


def test_is_defined():
    assert is_defined(None)
    assert not is_defined(UNDEFINED)
    assert not is_defined(Undefined)
    assert not is_defined(Undefined())


def test_default_if_undefined():
    assert default_if_undefined(UNDEFINED, 123) == 123
    assert default_if_undefined(None, 123) is None


def test_undefined_evaluates_as_false():
    assert bool(UNDEFINED) is False
    assert bool(Undefined) is False
    assert bool(Undefined()) is False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

def test_import_version():
    """ Test theat the package version is defined and that imports are working. """
    from ptutils.version import version
    assert isinstance(version, str)
    assert bool(version)



#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides unit tests for `ptutils.introspection`.
"""

from ptutils.introspection import get_function, get_functions


def dummy():
    pass


def dummy_a():
    pass


def dummy_b():
    pass


def dummy_c():
    pass


def test_get_function():
    assert get_function('dummy', __name__) is dummy
    assert get_function('does_not_exist', __name__) is None


def test_get_functions():
    f = get_functions(
        regex    = r'^dummy_([a-z]+)$',
        expander = None,
        module   = __name__
    )
    assert f == {
        "dummy_a": dummy_a,
        "dummy_b": dummy_b,
        "dummy_c": dummy_c,
    }
    assert len(f) == 3
    assert set(f.keys()) == set(['dummy_a', 'dummy_b', 'dummy_c'])
    assert f['dummy_a'] is dummy_a
    assert f['dummy_b'] is dummy_b
    assert f['dummy_c'] is dummy_c


def test_get_functions_with_expander():
    f = get_functions(
        regex    = r'^dummy_([a-z]+)$',
        expander = r'\1',
        module   = __name__
    )
    assert f == {
        "a": dummy_a,
        "b": dummy_b,
        "c": dummy_c,
    }
    assert len(f) == 3
    assert set(f.keys()) == set(['a', 'b', 'c'])
    assert f['a'] is dummy_a
    assert f['b'] is dummy_b
    assert f['c'] is dummy_c



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Simple decorator system for declaring CLI commands and their arguments.

This moduleis similar in concept to click, but with fewer features.
"""


import sys
import argparse
import functools
import datetime
from typing import Any, Callable, Dict, Iterable, Mapping, Optional, Tuple, Union

from ptutils.undefined import UNDEFINED, is_undefined, is_defined, default_if_undefined
from ptutils.typing import is_sequence, is_mapping
from ptutils.encoding import encode_json
from ptutils.punning import listify
from ptutils.logging import getLogger, set_verbosity, DEFAULT_VERBOSITY

# ------------------------------------------------------------------------------------------------------------------------
# Logging config
# ------------------------------------------------------------------------------------------------------------------------
logger = getLogger(__name__)


# ------------------------------------------------------------------------------------------------------------------------
# Globals
# ------------------------------------------------------------------------------------------------------------------------
""" The global registry of argument groups. """
__ARG_GROUPS__ = dict()

""" The global registry of command line commands. """
__COMMANDS__   = dict()


# ------------------------------------------------------------------------------------------------------------------------
# Utility functions
# ------------------------------------------------------------------------------------------------------------------------
def normalize_command_name(name: str) -> str:
    """
    Normalize a function name to produce the corresponding command name.

    Parameters
    ----------
    name : str
        The function name.

    Returns
    -------
    str
        The function name converted to lowecase with underscores replaced with hyphens
        and leading/trailing, or repeated hyphens removed.
    """
    name_  = name.lower().replace('_', '-')
    parts_ = [ x for x in name_.split('-') if x ]
    rv     = '-'.join( parts_ )
    return rv


# ------------------------------------------------------------------------------------------------------------------------
def _normalize_arg_name(name: str) -> str:
    """
    Return the normalized argument name of a given string.

    Parameters
    ----------
    name : str
        [description]

    Returns
    -------
    str
        The normalized name.

    Examples
    --------

    >>> _normalize_arg_name('-d')
    d
    >>> _normalize_arg_name('--debug-XyZ')
    debug_XyZ
    >>> _normalize_arg_name('---------a--b--c--------')
    a_b_c

    """

    name = name.lstrip('-').replace('-', '_')
    name = '_'.join([ x for x in name.split('_') if x])
    return name


# ------------------------------------------------------------------------------------------------------------------------
def _arg_name(args: Tuple[Any], kwargs: Dict[str, Any]) -> str:
    """
    Determine an argument's name based on specification
    (`argparse.ArgumentParser.add_argument` arguments).

    Parameters
    ----------
    args : Tuple[Any]
        The argument specification's positional arguments.
    kwargs : [type]
        The argument specification's keyword arguments.

    Returns
    -------
    str
        The resulting name (dest) of the argument.

    Raises
    ------
    Exception
        If no name can be determined.

    Notes
    -----
        This will produce the argument name by first looking for a 'dest' keyword
        argument. If not found, this will look at positional arguments beginning
        with '--' then those beginning with '-'. The positional arguments will have
        hyphens replaced with underscores, and leading/trailing/duplicate
        underscores removed.

    """

    name = kwargs.get('dest', UNDEFINED)
    if is_defined(name):
        return name

    # get positional arguments which may specify argument name (strings starting with hyphens)
    nd = [x for x in args if isinstance(x, str) and not x.startswith('-')]
    dd = [x for x in args if isinstance(x, str) and x.startswith('--')]
    sd = [x for x in args if isinstance(x, str) and x.startswith('-') and (x not in dd)]
    _d = nd + dd + sd

    # ensure there's at least one positional argument we could use to set the name.
    if not _d:
        raise Exception(
            "Can not determine shared argument name from supplied "
            f"arguments: args={args}, kwargs={kwargs}"
        )

    # select the first item in the list, normalize the name and return it.
    return _normalize_arg_name(_d[0])


# ------------------------------------------------------------------------------------------------------------------------
def _resolve(subject: Union[str, Callable]) -> str:
    """
    Given a string or a function, determine the name the cli simple system will use for lookups
    for shared arguments and argument groups.

    Parameters
    ----------
    subject : Union[str, Callable]
        A string or function.

    Returns
    -------
    str
        Either the string unmodified, or in the case of a callable, the callable's __name__.

    Raises
    ------
    TypeError
        When `subject` is not a string and not callable.
    """
    if callable(subject):
        return subject.__name__
    elif isinstance(subject, str):
        return subject
    else:
        raise TypeError(
            f"Object of type '{type(subject).__name__}' may not "
            "be used to reference a command, arg, arg_group, or shared_arg."
        )


# ------------------------------------------------------------------------------------------------------------------------
def _resolve_command(subject: Union[str, Callable]) -> str:
    """
    Given a string or a function, determine the name the cli simple system will use for lookups
    for commands.

    Parameters
    ----------
    subject : Union[str, Callable]
        A string or function.

    Returns
    -------
    str
        Either the string unmodified, or in the case of a callable, the callable's __name__.

    Raises
    ------
    TypeError
        When `subject` is not a string and not callable.
    """
    return normalize_command_name(_resolve(subject))


# ------------------------------------------------------------------------------------------------------------------------
def get_command(name: str) -> Callable:
    """
    Lookup a registered command.

    Parameters
    ----------
    name : str
        The command or function name. Will be normalized with `normalize_command_name` prior to searching.

    Returns
    -------
    Callable
        The function corresponding to that command, if found.

    Raises
    ------
    Exception
        When no such command name has been registered.
    """
    name_ = _resolve_command(name)
    if name_ not in __COMMANDS__:
        raise Exception(f"Unknown command: No command has been defined with name '{name_}'.")
    return __COMMANDS__[name_]


# ------------------------------------------------------------------------------------------------------------------------
def _register_arg_group(name: str, func: Callable):
    """
    Utility function to initialize an argument group on a function and register it globally.

    Parameters
    ----------
    name : atr
        The name of the argument group to add.
    func : Callable
        The function to be decorated as an argument group.

    Raises
    ------
    Exception
        If the argument group is already defined.
    Exception
        If the function is an already decorated command or argument group.
    """
    if name in __ARG_GROUPS__:
        raise Exception(f"Argument group '{name}' is already defined.")
    if hasattr(func, '__cmd_info__'):
        raise Exception(f"Registering argument group {name} on function {func.__name__} which already has metadata.")
    func.__cmd_info__ = dict(
        arg_info   = [],
        call       = lambda kwargs: func(**kwargs)
    )

    __ARG_GROUPS__[name] = func


# ------------------------------------------------------------------------------------------------------------------------
def arg_group_exists(name: Union[str, Callable]) -> bool:
    """
    Check if the named argument group exists

    Parameters
    ----------
    name : Union[str, Callable]
        The name or decorated function of the argument group

    Returns
    -------
    bool
        True if the argument group is defined.
    """
    return _resolve(name) in __ARG_GROUPS__


# ------------------------------------------------------------------------------------------------------------------------
def command_exists(name: Union[str, Callable]) -> bool:
    """
    Check if the named command exists

    Parameters
    ----------
    name : str
        The command or function name, or decorated function itself. Will be normalized
        with `normalize_command_name` prior to searching.

    Returns
    -------
    bool
        True if the command is defined.
    """
    return _resolve_command(name) in __COMMANDS__


# ------------------------------------------------------------------------------------------------------------------------
def shared_arg_exists(name: Union[str, Callable]) -> bool:
    """
    Check if the named shared argument exists

    Parameters
    ----------
    name : Union[str, Callable]
        The name of the shared argument or the return value of `shared_arg`.

    Returns
    -------
    bool
        True if the argument is defined.

    Notes
    -----
    A `shared_arg` is just a single-argument argument group of the same name. As
    such this will ALSO return True if there's an argument group by that name.
    """
    return _resolve(name) in __ARG_GROUPS__


# ------------------------------------------------------------------------------------------------------------------------
def remove_shared_arg(name: Union[str, Callable]) -> None:
    """
    Removes the shared argument definition from the global registry.
    This is useful in unit testing.

    Parameters
    ----------
    name : Union[str, Callable]
        The name of the shared argument or the return value of `shared_arg`.

    Notes
    -----
    This implementation does not remove any references to the shared
    argument, and so may entail undefined behavior if such references are
    not also cleaned up.
    """
    name_ = _resolve(name)
    if name_ in __ARG_GROUPS__:
        del __ARG_GROUPS__[name_]
    else:
        raise Exception(f"No shared arg or arg group by name '{name_}', so nothing to delete.")


# ------------------------------------------------------------------------------------------------------------------------
def remove_arg_group(name: Union[str, Callable]) -> None:
    """
    Removes the argument group definition from the global registry.
    This is useful in unit testing.

    Parameters
    ----------
    name : Union[str, Callable]
        The name or decorated function of the argument group.

    Notes
    -----
    This implementation does not remove any references to the argument
    group, and so may entail undefined behavior if such references are
    not also cleaned up.
    """
    remove_shared_arg(name)


# ------------------------------------------------------------------------------------------------------------------------
def remove_command(name: Union[str, Callable]) -> None:
    """
    Remove a command from the global list of registered commands.

    Parameters
    ----------
    name : Union[str, Callable]
        The command or function name, or decorated function itself. Will be normalized
        with `normalize_command_name` prior to searching.

    """
    name_ = _resolve_command(name)
    if name_ not in __COMMANDS__:
        raise Exception(f"No such command registered: '{name}'")
    print(f"Removing command: '{name}' ('{name_}')")
    del __COMMANDS__[name_]


# ------------------------------------------------------------------------------------------------------------------------
def get_argument(name: Union[str, Callable], argument_name: str) -> Optional[Mapping[ str, Any ]]:
    """
    Given a command or an argument group, return the specification of the named argument.

    Parameters
    ----------
    name : Union[str, Callable]
        The name or decorated function for a command or argument group.
    argument_name : str
        The argument name

    Returns
    -------
    Optional[Mapping[ str, Any ]]
        The argument spec, or None if no argument by that name was found in the argument list.

    Raises
    ------
    ValueError
        If `name` does not refer to a registered command or argument group.

    """
    name_ = _resolve(name)
    if name_ in __ARG_GROUPS__:
        func = __ARG_GROUPS__[name_]
    else:
        name_ = _resolve_command(name)
        if name_ in __COMMANDS__:
            func = __COMMANDS__[name_]
        else:
            raise ValueError(f"could not resolve '{name}' asa known command or argument group.")

    _argument_name = _resolve(argument_name)
    for spec in _walk_argument_specs(func):
        if spec['name'] == _argument_name:
            return spec

    return None


# ------------------------------------------------------------------------------------------------------------------------
def has_argument(name: Union[str, Callable], argument_name: str) -> bool:
    """
    Check if a command or argument group includes an argument with the supplied name.

    Parameters
    ----------
    name : Union[str, Callable]
        [description]
    argument_name : str
        [description]

    Returns
    -------
    bool
        [description]
    """
    return get_argument(name, argument_name) is not None


# ------------------------------------------------------------------------------------------------------------------------
def _walk_argument_specs(
    subject:       Callable,
    _group_chain:  Optional[ Iterable[ str ] ] = None
) -> Tuple[Mapping[ str, Any ], Iterable[ str ]]:
    """
    Iterate over the argument specifications for a command, argument group, or shared argument.

    Parameters
    ----------
    subparser : argparse.ArgumentParser
        The subparser for the command.
    arg_info : Union[ Iterable[ Mapping[ str, Any ] ], Mapping[ str, Any ] ]
        Argument definition metadata
    _group_chain : Optional[ Iterable[ str ] ], optional
        Recursion helper to help properly reporting errors due to usage of shared args or arg groups, by default None.

    Raises
    ------
    Exception
        When an unknown argument or argument group is referenced.
    TypeError
        When an argument specification is of an unknown object type.
    TypeError
        When an argument specification is of an invalid specification type.
    """

    # initialize the gorup chain
    _group_chain = _group_chain or []

    # ensure subject is a command or argument group
    if (
        (not callable(subject)) or
        (not hasattr(subject, '__cmd_info__'))
    ):
        raise ValueError(
            f"Object of type '{type(subject).__name__}' may not be "
            "used as a command, argument group, or shared argument."
            "Did you forget an @command, @arg_group decorator?"
        )

    # get metadata
    __cmd_info__ = subject.__cmd_info__

    # ensure argument info is available.
    if 'arg_info' not in __cmd_info__:
        raise ValueError(
            f"Argument information missing on object of type '{type(subject).__name__}."
            "Did you forget an @command, @arg_group decorator?"
        )

    # get argument info
    specs = listify(__cmd_info__['arg_info'])

    # walk the list
    for spec in specs:
        if not is_mapping(spec):
            raise RuntimeError(
                f"Object of type '{type(spec).__name__}' is "
                "not a valid argument specification."
            )

        kind = spec.get('kind', UNDEFINED)
        if kind == "arg":
            # an argument definition. yield it directly.
            yield spec

        elif kind == 'arg_group':
            # an usage definition. recurse into the named group and yield all children.

            name = spec.get('name', UNDEFINED)
            if is_undefined(name):
                raise Exception("Invalid argument group specification: missing name.")
            if name not in __ARG_GROUPS__:
                raise Exception(
                    f"Invalid argument group specification: Name '{name}' is "
                    f"not a registered argument group: {list(__ARG_GROUPS__.keys())}."
                )

            child_subject     = __ARG_GROUPS__[name]
            child_group_chain = _group_chain + [name]
            for child_spec in _walk_argument_specs(
                subject      = child_subject,
                _group_chain = child_group_chain
            ):
                yield child_spec
        else:
            raise ValueError(f"Unknown kind of argument spec: '{kind}'")


# ------------------------------------------------------------------------------------------------------------------------
# Decorators
# ------------------------------------------------------------------------------------------------------------------------
def command(func: Callable[..., None]) -> Callable[..., None]:
    """
    Decorator for marking a function as a command.

    Parameters
    ----------
    func : Callable[...,None]
        The command implementation.

    Returns
    -------
    Callable[...,None]
        The command, instrumented for calling form the command line with any
        agruments, if defined.

    Raises
    ------
    Exception
        If a command by that name is already registered, or is an argument group or
        shared argument.
    """
    name = normalize_command_name(func.__name__)

    if name in __COMMANDS__:
        raise Exception(f"Command '{name}' is already defined.")

    __COMMANDS__[name] = func

    if hasattr(func, '__cmd_info__'):
        raise Exception(f"Command '{name}' already has metadata.")
    func.__cmd_info__ = dict(
        arg_info   = [],
        call       = lambda kwargs: func(**kwargs)
    )

    return func

# def instance_command(func: Callable[...,None])->Callable[...,None]:
#     name = normalize_command_name(func.__name__)

#     if name in __COMMANDS__:
#         raise Exception(f"Command '{name}' is already defined.")

#     __COMMANDS__[name] = func

#     if hasattr(func, '__cmd_info__'):
#         raise Exception(f"Command '{name}' already has metadata.")

#     func.__cmd_info__ = dict(
#         arg_info   = [],
#         call       = lambda **kwargs: call_instance_command(func, **kwargs)
#     )

#     @classmethod
#     @functools.wraps(func)
#     def _wrapped_instance_command(cls, **kwargs):
#         instance = cls(**kwargs)
#         return func(instance)

#     return func


# ------------------------------------------------------------------------------------------------------------------------
def arg(*args, **kwargs) -> Callable:
    """
    Parametric decorator defines a command argument.

    Parameters
    ----------
    *args: Iterable[Any]
        Positional arguments supplied to `argparse.ArgumentParser.add_argument` when
        adding this argument.
    **kwargs:
        Keyword arguments supplied to `argparse.ArgumentParser.add_argument` when
        adding this argument.

    Returns
    -------
    Callable
        The command function with argument definition metadata attached.

    Raises
    ------
    Exception
        When applied to a function which is not a decorated command, or if the argument
        name is already registered.
    """
    def decorate(func: Callable) -> Callable:
        if not hasattr(func, '__cmd_info__'):
            raise Exception(
                f"Function '{func.__name__}' is not a command. "
                "Did you forget the '@command' decorator?"
            )

        spec = {
            "kind":   "arg",
            "args":   args,
            "kwargs": kwargs,
            "name":   _arg_name(args, kwargs)
        }
        func.__cmd_info__['arg_info'].append(spec)
        return func
    return decorate


# ------------------------------------------------------------------------------------------------------------------------
def use_arg_group(name: Union[str, Callable]) -> Callable:
    """
    Parametric decorator that adds a group of arguments to a command.

    Parameters
    ----------
    name : str
        The name of the argument group (defined elswhere with `arg_group` decorator).

    Returns
    -------
    Callable
        A decorator which injects the collected argument definition to the decorated
        function's metadata.
    """

    def decorate(func: Callable) -> Callable:
        """
        Internal decorator for `use_arg_group` that injects argument group usage
        details into decorated command metadata.

        Parameters
        ----------
        func : Callable
            The decorated function

        Returns
        -------
        Callable
            The function with argument group metadata injected.
        """
        name_ = _resolve(name)
        if name_ not in __ARG_GROUPS__:
            raise Exception(f"Unknown argument group or shared argument: '{name}'")
        spec = {
            "kind":   "arg_group",
            "name":   name_
        }

        func.__cmd_info__['arg_info'].append( spec )
        return func
    return decorate


# ------------------------------------------------------------------------------------------------------------------------
def use_shared_arg(name: str) -> Callable:
    """
    Parametric decorator that allows sharing of argument defined elsewhere with `shared_arg`.

    Parameters
    ----------
    name : str
        The name of the argument.

    Returns
    -------
    Callable
        A decorator which injects the argument definition to the decorated
        function's metadata.
    """
    return use_arg_group(name)


# ------------------------------------------------------------------------------------------------------------------------
def arg_group(func):
    """
    Decorator which defines an argument group.

    Parameters
    ----------
    func : [type]
        [description]

    Returns
    -------
    Callable
        The decorated function with argument metadata injected.

    Raises
    ------
    Exception
        If the decorated function is already a command or argument group, or if the
        argument group name matches an already defined shared argument name.

    Notes
    -----
    The decorated function should have an empty body.
    Calling the decorated function directly may yield undefined behavior.

    Examples
    --------

    >>> @arg('-n', '--number', dest="The number of the counting?", metavar="COUNT", default=3, type=int)
    ... @arg_group
    ... def thing_counts(): pass
    >>> @use_arg_group('thing_counts')
    ... @command
    ... def count_things( thing_counts: int) -> int:
    ...    return thing_counts

    """

    name = func.__name__

    @functools.wraps(func)
    def __cli_internal__(**kwargs):
        raise Exception("An arg group is not callable")

    _register_arg_group(name, __cli_internal__)
    return __cli_internal__


# ------------------------------------------------------------------------------------------------------------------------
# Non-decorator declarations
# ------------------------------------------------------------------------------------------------------------------------
def shared_arg(*args, **kwargs) -> None:
    """
    Define a shared argument.

    Parameters
    ----------
    *args: Iterable[Any]
        Positional arguments supplied to `argparse.ArgumentParser.add_argument` when adding this argument.
    **kwargs:
        Keyword arguments supplied to `argparse.ArgumentParser.add_argument` when adding this argument.

    Raises
    ------
    Exception
        If the argument name can not be determined.
    Exception
        If an argument or argument group by that name already exists.
    """

    name = _arg_name(args, kwargs)

    def __cli_internal__():
        raise Exception("An arg group is not callable")
    __cli_internal__.__name__ = name

    _register_arg_group(name, __cli_internal__)
    spec = {
        "kind":   "arg",
        "args":   args,
        "kwargs": kwargs,
        "name":   name
    }
    __cli_internal__.__cmd_info__['arg_info'].append(spec)

    return __cli_internal__


# ------------------------------------------------------------------------------------------------------------------------
# Entrypoint utility functions
# ------------------------------------------------------------------------------------------------------------------------
def add_command_subparser_args(
    subparser:     argparse.ArgumentParser,
    arg_info:      Union[ Iterable[ Mapping[ str, Any ] ], Mapping[ str, Any ] ],
    _group_chain:  Optional[ Iterable[ str ] ] = None
) -> None:
    """
    Add arguments for a command's subparser based on metadata attached to the decorated function.

    Parameters
    ----------
    subparser : argparse.ArgumentParser
        The subparser for the command.
    arg_info : Union[ Iterable[ Mapping[ str, Any ] ], Mapping[ str, Any ] ]
        Argument definition metadata
    _group_chain : Optional[ Iterable[ str ] ], optional
        Recursion helper to help properly reporting errors due to usage of shared args or arg groups, by default None.

    Raises
    ------
    Exception
        When an unknown argument or argument group is referenced.
    TypeError
        When an argument specification is of an unknown object type.
    TypeError
        When an argument specification is of an invalid specification type.
    """
    _group_chain = _group_chain or []
    if is_sequence(arg_info):
        for spec in arg_info:
            add_command_subparser_args(
                subparser,
                arg_info     = spec,
                _group_chain = _group_chain
            )
    elif is_mapping(arg_info):
        kind = arg_info['kind']
        if kind == "arg_group":
            name = arg_info['name']
            if name not in __ARG_GROUPS__:  # pragma: no cover
                cmd = subparser.get_default('command_name')
                gcs = ' -> '.join(_group_chain)
                gcm = '' if not gcs else f" indirectly via group(s): {gcs}"
                raise Exception(f"Unknown argument group '{name}' referenced by command '{cmd}'{gcm}.")
            add_command_subparser_args(
                subparser,
                arg_info     = __ARG_GROUPS__[name].__cmd_info__['arg_info'],
                _group_chain = _group_chain + [name]
            )
        elif kind == "arg":
            args   = arg_info['args']
            kwargs = arg_info['kwargs']
            subparser.add_argument(*args, **kwargs)
        else:  # pragma: no cover
            raise Exception(f"Unknown argument spec type: '{kind}'")
    else:  # pragma: no cover
        raise TypeError(f"Invalid type for argument spec: '{type(arg_info).__name__}'")


# ------------------------------------------------------------------------------------------------------------------------
""" The main argument parser. """
__PARSER__ = None


# ------------------------------------------------------------------------------------------------------------------------
def program_defaults(
    program:           Optional[str] = None,
    version:           Optional[str] = None,
    copyright_message: Optional[str] = None,
) -> Tuple[str, str, str]:
    """
    Apply default values for program, version, and copyright message to be included in cli
    application help text.

    Parameters
    ----------
    program : str, optional
        The name of the program. Defaults to `sys.argv[0]`
    version : str
        The current version of the program. Defaults to 'v?.?.?'
    copyright_message : Optional[str]
        The copyright message for the program. Defaults to '(C) copyright YEAR. All rights reserved)'

    Returns
    -------
    Tuple[str, str, str]
        The program, version, and copyright_message in a tuple in respective order.
    """
    year = datetime.datetime.utcnow().year
    if copyright_message is None:
        copyright_message = f"(C) Copyright {year}. All rights reserved."
    if version is None:
        version = "v?.?.?"
    if program is None:
        program = sys.argv[0]

    return (program, version, copyright_message)


# ------------------------------------------------------------------------------------------------------------------------
def get_args_parser(
    program:           Optional[str] = None,
    version:           Optional[str] = None,
    copyright_message: Optional[str] = None
) -> argparse.ArgumentParser:
    """
    Get the global argument parser for all registered commands and associated arguments.

    Parameters
    ----------
    program : str, optional
        The name of the program. Defaults to `sys.argv[0]`
    version : str
        The current version of the program. Defaults to 'v?.?.?'
    copyright_message : Optional[str]
        The copyright message for the program. Defaults to '(C) copyright YEAR. All rights reserved)'

    Returns
    -------
    argparse.ArgumentParser
        An argument parser which handles all commands.

    """
    global __PARSER__
    if __PARSER__ is None:

        # get the program default values
        (program, version, copyright_message) = program_defaults(
            program, version, copyright_message
        )

        # initialize the main parser
        parser = argparse.ArgumentParser(
            prog = sys.argv[0],
            description = (f"{program} v{version}"),
            epilog = copyright_message or "",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            add_help=False
        )
        parser.add_argument(
            '-h', '--help',
            help    = 'Display this help information.',
            dest    = 'help',
            action  = "store_true"
        )
        subparsers = parser.add_subparsers(
            title       = 'Commands',
            description = (
                f'For help a command, type:\n\t'
                f'{sys.argv[0]} COMMAND --help.\n'
                'The following commands are supported:'
            )
        )

        for (command_name, command_function) in __COMMANDS__.items():
            # print(f"Adding the {command_name} command: {command_function.__name__}(...)")
            description  = command_function.__doc__ or f'The {command_name} command.'
            subparser = subparsers.add_parser(
                command_name,
                description = description,
                help        = description
            )
            subparser.set_defaults(
                command_name     = command_name,
                command_function = command_function,
                subparser        = subparser
            )
            subparser.add_argument(
                '-v', '--verbose',
                help    = 'Increase logging verbosity.',
                dest    = 'verbosity',
                action  = "count"
            )
            subparser.add_argument(
                '-q', '--quiet',
                help    = 'Decrease logging verbosity.',
                dest    = '__quiet',
                action  = "count"
            )
            # subparser.add_argument(
            #     '-h', '--help',
            #     help    = f"Display help for the '{command_name}' command.",
            #     dest    = 'help',
            #     action  = "store_true"
            # )
            add_command_subparser_args(
                subparser,
                command_function.__cmd_info__['arg_info']
            )

        __PARSER__ = parser

    return __PARSER__


# ------------------------------------------------------------------------------------------------------------------------
def convert_undefined_attributes_to_none( args: object ) -> None:
    """
    Change any `UNDEFINED` attributes to `None`.
    This will be modified in-place, and nothing is returned.

    Parameters
    ----------
    args : object
        An arguments object to modify.
    """
    # Convert any undefined values to None
    for (k, v) in args.__dict__.items():
        if is_undefined(v):
            args.__dict__[k] = None


# ------------------------------------------------------------------------------------------------------------------------
def apply_verbosity_args( args: object ) -> None:
    """
    Handle verbosity and quiet arguments, setting log verbosity accordingly.
    The 'verbosity' and '__quiet' arguments (which were automatically added)
    will be captured, applied, and removed from the `args` object, in-place.

    Parameters
    ----------
    args : object
        An arguments object.
    """
    # set log verbosity
    verbosity = getattr(args, 'verbosity', UNDEFINED)
    if is_defined(verbosity):  # pragma : no cover
        delattr(args, 'verbosity')
    verbosity = default_if_undefined(verbosity, 0)
    verbosity = verbosity if (verbosity is not None) else 0

    quiet     = getattr(args, '__quiet', UNDEFINED)
    if is_defined(quiet):  # pragma : no cover
        delattr(args, '__quiet')
    quiet = default_if_undefined(quiet, 0)
    quiet = quiet if (quiet is not None) else 0

    set_verbosity( DEFAULT_VERBOSITY + verbosity - quiet )


# ------------------------------------------------------------------------------------------------------------------------
def get_args(
    args:              Optional[Iterable[str]] = None,
    program:           Optional[str] = None,
    version:           Optional[str] = None,
    copyright_message: Optional[str] = None,
) -> Tuple[object, argparse.ArgumentParser]:
    """
    Get an argument parser based on all registered `command` functions, parse supplied
    arguments, and return an arguments object.

    Parameters
    ----------
    args: Iterable[str], optional
        A list of positional arguments excluding the program name, defaults to `sys.argv[1:]`
    program : str, optional
        The name of the program. Defaults to `sys.argv[0]`
    version : str
        The current version of the program. Defaults to 'v?.?.?'
    copyright_message : Optional[str]
        The copyright message for the program. Defaults to '(C) copyright YEAR. All rights reserved)'


    Returns
    -------
    Tuple[object, argparse.ArgumentParser]
        A tuple whose first element is an arguments object ( as returned by
        `argparse.ArgumentParser.parse_arguments` ) with attributes matching the registered
        argument names for the selected command, and whose second element is the argument
        parser itself.

    """
    # get the program default values
    (program, version, copyright_message) = program_defaults(
        program, version, copyright_message
    )
    if args is None:
        args = sys.argv[1:]

    # get parser
    try:
        parser = get_args_parser(
            program = program,
            version = version,
            copyright_message = copyright_message
        )
    except Exception as e:
        raise Exception("Error building parser(s).") from e

    # parse args
    try:
        args = parser.parse_args( args = args )
    except SystemExit:
        raise
    except Exception as e:  # noqa F722  # pragma : no cover
        raise Exception("Error parsing arguments.") from e

    # Convert any undefined values to None
    convert_undefined_attributes_to_none(args)

    # determine if we have a specified command
    subparser = getattr(args, 'subparser', None)
    if subparser is None:
        logger.error("No command specified")
        parser.print_help(file=sys.stderr)
        sys.exit(-1)
    delattr(args, 'subparser')

    # done. return the result
    return args, subparser


# ------------------------------------------------------------------------------------------------------------------------
def get_command_from_args(args: object) -> Tuple[Callable, str]:
    """
    Given a parsed arguments object, extract the command function and name, and
    remove those attributes from the args object, in-place.

    Parameters
    ----------
    args : object
        An arguments object.

    Returns
    -------
    Tuple[Callable, str]
        The command function and command name.
    """
    command      = args.command_function
    command_name = args.command_name
    delattr(args, 'command_function')
    delattr(args, 'command_name')
    return (command, command_name)


# ------------------------------------------------------------------------------------------------------------------------
def handle_help_flag(args: object, parser: argparse.ArgumentParser) -> None:
    """
    Explicitly handle displaying of help information for a command when
    the --help flag is specified. When present and True, the help is displayed
    and the program exits. When present and False, the help attribute is
    removed from args, in-place. When absent, no action is taken.

    Parameters
    ----------
    args : object
        An arguments object
    parser : argparse.ArgumentParser
        The argument parser.

    """
    help = getattr(args, 'help', None)
    if help is not None:
        if help:
            parser.print_help(file = sys.stderr)
            return 0
        delattr(args, 'help')


# ------------------------------------------------------------------------------------------------------------------------
def display_invocation_details(command_name: str, kwargs: Dict[str, Any]) -> None:
    """
    Display program invocation details.

    Parameters
    ----------
    command_name : str
        The selected command name.
    kwargs : Dict[str, Any]
        Parsed dictionary of command line arguments.
    """
    logger.debug('=======================================================')
    logger.debug('Command:   %s' % command_name)

    if kwargs:
        logger.debug('Arguments:')
        mk = max(len(k) for k in kwargs) + 1
        for (k, v) in kwargs.items():
            try:
                val = encode_json(v)
            except:  # noqa F722
                val = repr(v)
            logger.debug(f"  {(k+':'):{mk}} {val}")
    else:
        logger.debug('Arguments: <<< No arguments provided >>>')
    logger.debug('=======================================================')


# ------------------------------------------------------------------------------------------------------------------------
def run_command(
    args:              Optional[Iterable[str]] = None,
    program:           Optional[str] = None,
    version:           Optional[str] = None,
    copyright_message: Optional[str] = None,
) -> int:
    """
    Parse supplied command line arguments and run the corresponding command if specified.

    Parameters
    ----------
    args: Iterable[str], optional
        A list of positional arguments excluding the program name, defaults to `sys.argv[1:]`
    program : str, optional
        The name of the program. Defaults to `sys.argv[0]`
    version : str
        The current version of the program. Defaults to 'v?.?.?'
    copyright_message : Optional[str]
        The copyright message for the program. Defaults to '(C) copyright YEAR. All rights reserved)'

    Returns
    -------
    int
        The command's exit code.
    """

    try:
        args, subparser = get_args(
            args              = args,
            program           = program,
            version           = version,
            copyright_message = copyright_message
        )
    except SystemExit:
        raise
    except:  # noqa F722  # pragma : no cover
        logger.exception("Error getting program arguments")
        return -1

    # apply logging verbosity args (verbose and quiet)
    apply_verbosity_args(args)

    # get command details
    command, command_name = get_command_from_args(args)

    # get help flag
    handle_help_flag(args, subparser)

    # convert args to a dictionary
    kwargs = vars(args)

    # log command invocation arguments
    display_invocation_details(command_name, kwargs)

    # invoke the command
    try:
        rv = command.__cmd_info__['call'](kwargs)
        if rv is None:
            return 0
        return rv
    except Exception as e:
        logger.exception(f"Error while running command '{command_name}': {e}")
        subparser.print_help()
        return -1



#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides unit tests for `ptutils.numerical`.
"""

from ptutils.numerical import __NAN__
from math import isnan


def test_nan():
    assert isnan(__NAN__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This module provides unit tests for `ptutils.encoding`."""

import pytest
import datetime
from ptutils.encoding import (
    HAVE_YAML,
    encode_json,
    encode_yaml,
    pretty_json,
    pretty_yaml,
    decode_json,
    decode_yaml,
    json_serial
)


def test_json_serial_handles_datetime():
    val = json_serial(datetime.datetime.utcnow())
    assert isinstance(val, str)
    assert bool(val)


def test_json_serial_handles_timedelta():
    val = json_serial(datetime.timedelta(seconds=5))
    assert isinstance(val, (int, float))
    assert val == 5


def test_json_serial_fails_with_unknown_class():
    class C:
        pass
    with pytest.raises(TypeError) as _:
        _ = json_serial(C())


NATIVE_CASES = [
    None,
    "string",
    123456,
    123456.789,
    True,
    False,
    [],
    [None],
    ["a", 2, 3.0, True, False, {}],
    {"a": 1, "5": 123, "p": False}
]


@pytest.mark.parametrize(
    "test_case",
    NATIVE_CASES,
    ids = [f"NATIVE_CASES[{i}]" for i in range(len(NATIVE_CASES))]
)
def test_json_roundtrip(test_case):
    encoded_test_case = encode_json(test_case)
    assert isinstance(encoded_test_case, str)
    assert bool(encoded_test_case)
    decoded_test_case = decode_json(encoded_test_case)
    assert decoded_test_case == test_case


@pytest.mark.parametrize(
    "test_case",
    NATIVE_CASES,
    ids = [f"NATIVE_CASES[{i}]" for i in range(len(NATIVE_CASES))]
)
def test_json_pretty_roundtrip(test_case):
    encoded_test_case = pretty_json(test_case)
    assert isinstance(encoded_test_case, str)
    assert bool(encoded_test_case)
    decoded_test_case = decode_yaml(encoded_test_case)
    assert decoded_test_case == test_case


@pytest.mark.parametrize(
    "test_case",
    NATIVE_CASES,
    ids = [f"NATIVE_CASES[{i}]" for i in range(len(NATIVE_CASES))]
)
def test_yaml_roundtrip(test_case):
    if HAVE_YAML:
        encoded_test_case = encode_yaml(test_case)
        assert isinstance(encoded_test_case, str)
        assert bool(encoded_test_case)
        decoded_test_case = decode_yaml(encoded_test_case)
        assert decoded_test_case == test_case
    else:
        pytest.skip("No YAML support")


@pytest.mark.parametrize(
    "test_case",
    NATIVE_CASES,
    ids = [f"NATIVE_CASES[{i}]" for i in range(len(NATIVE_CASES))]
)
def test_yaml_pretty_roundtrip(test_case):
    if HAVE_YAML:
        encoded_test_case = pretty_yaml(test_case)
        assert isinstance(encoded_test_case, str)
        assert bool(encoded_test_case)
        decoded_test_case = decode_yaml(encoded_test_case)
        assert decoded_test_case == test_case
    else:
        pytest.skip("No YAML support")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This module provides unit tests for `ptutils.debug`."""

from ptutils.program import Program


def test_program():
    p = Program("echo hello")
    p.start()
    rc = p.wait()
    assert rc == 0
    assert p.stdout == "hello"
    assert p.stderr == ""



#!/bin/false
# -*- coding: utf-8 -*-

"""This module provides unit tests for `ptutils.globbing`."""

import os
import pytest
import tempfile
import shutil

FILE_STRUCTURE = {
    "boring_things": {
        "thing.json":      '{ "a": 1, "b": true, "c": 123.45, "d": "hello world" }',
        "thing.yaml":      '---\na: 1\nb: true\nc: 123.45\nd: hello world',
        "text_thing":      'this is file 2',
        "tbd_things": {
        }
    },
    "interesting_things": {
        "scary_thing": "Dracula",
        "silly_thing": "Spongebob",
        "bright_thing": "The sun",
        "heavy_things": {
            "divorce": "nope",
            "lead": "plumbum"
        },
        "auto_things": {
            "mobile": "car",
            "graph": "signature"
        },
        "missing_things": {}
    },
    "apples": {
        "granny-smith": "green",
        "red-delicious": "red"
    },
    "a_file": "some content",
    "another_file": "some other content"
}
PATH_THINGS = ""
PATH_BORING_THINGS = "/boring_things"
PATH_TBD_THINGS = "/boring_things/tbd_things"
PATH_INTERESTING_THINGS = "/interesting_things"
PATH_HEAVY_THINGS = "/interesting_things/heavy_things"
PATH_AUTO_THINGS_THINGS = "/interesting_things/auto_things"
PATH_MISSING_THINGS = "/interesting_things/missing_things"
PATH_APPLES = "apples"

THING_SUBDIRS = ["/boring_things", "/interesting_things", "/apples"]
THING_SUBFILES = ["/a_file", "/another_file"]
THING_SUBS = THING_SUBDIRS + THING_SUBFILES

THING_SUBDIRS_REC = [
    PATH_BORING_THINGS,
    PATH_TBD_THINGS,
    PATH_INTERESTING_THINGS,
    PATH_HEAVY_THINGS,
    PATH_AUTO_THINGS_THINGS,
    PATH_MISSING_THINGS,
    PATH_APPLES
]
THING_SUBFILES_REC = [
    "/a_file",
    "/another_file",
    PATH_BORING_THINGS + "/thing.json",
    PATH_BORING_THINGS + "/thing.yaml",
    PATH_BORING_THINGS + "/text_thing",
    PATH_INTERESTING_THINGS + "/scary_thing",
    PATH_INTERESTING_THINGS + "/silly_thing",
    PATH_INTERESTING_THINGS + "/bright_thing",
    PATH_HEAVY_THINGS + "/divorce",
    PATH_HEAVY_THINGS + "/lead",
    PATH_AUTO_THINGS_THINGS + "/mobile",
    PATH_AUTO_THINGS_THINGS + "/graph",
    PATH_APPLES + "/granny-smith",
    PATH_APPLES + "/red-delicious",
]

THING_KEYED_SUBFILES_REC = [
    PATH_BORING_THINGS + "/text_thing",
    PATH_INTERESTING_THINGS + "/scary_thing",
    PATH_INTERESTING_THINGS + "/silly_thing",
    PATH_INTERESTING_THINGS + "/bright_thing",
]

STARTS_WITH_A_REC = [
    "/a_file",
    "/another_file",
    PATH_AUTO_THINGS_THINGS,
    PATH_APPLES
]


def path_parts(path: str):
    return [x for x in path.split('/') if x ]


def localize_path(path: str):
    return path.replace('/', os.sep)


def populate(path, content):
    if isinstance(content, dict):
        os.makedirs(path, exist_ok=True)
        for (k, v) in content.items():
            populate(os.path.join(path, k), v)
    elif isinstance(content, str):
        with open(path, "w") as filp:
            filp.write(content)
    else:
        raise Exception("BUG: content is not a string and is not a dict.")


@pytest.fixture(autouse=True)
def sample_file_structure():
    dir = tempfile.TemporaryDirectory()
    populate(dir.name, FILE_STRUCTURE)
    yield dir.name
    shutil.rmtree(dir.name)


""" Queries for globbing and expected results. """
GLOBBING_QUERIES = [
    # (path, pattern, recursive, filter_func, expected_relpaths)
    (PATH_THINGS, None, False, None,           THING_SUBS),
    (PATH_THINGS, None, False, os.path.isfile, THING_SUBFILES),
    (PATH_THINGS, None, False, os.path.isdir,  THING_SUBDIRS),
    (PATH_THINGS, None, True,  os.path.isfile, THING_SUBFILES_REC),
    (PATH_THINGS, None, True, os.path.isdir,  THING_SUBDIRS_REC),
    (PATH_THINGS, r'^.*_thing$', True, None,  THING_KEYED_SUBFILES_REC),
    (PATH_THINGS, r'^a.*$', True, None,  STARTS_WITH_A_REC),
]



#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides functions and classes for dealing with time and timing.
"""

import datetime
from typing import Optional


def now() -> datetime.datetime:
    """ Get current time as UTC datetime.

    Returns
    -------
    now: datetime.datetime:
        The current UTC date and time.

    """
    return datetime.datetime.utcnow()


def dt(start: datetime.datetime, end: Optional[datetime.datetime] = None) -> datetime.timedelta:
    """ Compute the elapsed time between a given start time and end time.

    Parameters
    ----------
    start: datetime.datetime
        The starting point in time.
    end: datetime.datetime, optional
        The ending point in time. If omitted, the current time will be used.


    Returns
    -------
    elapsed_time: datetime.timedelta:
        The elapsed time.

    """
    end = end if (end is not None) else now()
    return end - start


# Import protection
if __name__ == '__main__':
    raise Exception(
        "This module is not executable. To use it, add "
        "'from ptutils.time import ...' to your python "
        "source code. Refer to the API documentation for more "
        "information."
    )



#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides unit tests for `ptutils.text`.
"""
import pytest
from ptutils.text import strip_quotes, strip_brackets, strip_line_ending


STRIP_QUOTES_TEST_CASES = [
    (b"'''hello'''", b"'''hello'''"),
    ("'hello'", "hello"),
    ('"hello"', "hello"),
    ('4', "4"),
    ('"4', '"4'),
    ("'4", "'4"),
]


@pytest.mark.parametrize(
    argnames  = ["obj", "expected"],
    argvalues = STRIP_QUOTES_TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[1])}" for x in STRIP_QUOTES_TEST_CASES]
)
def test_strip_quotes(obj, expected):
    assert strip_quotes(obj) == expected


SQUARE_BRACKETS = [['[', ']']]
CURLY_BRACKETS  = [['{', '}']]
PAREN_BRACKETS  = [['(', ')']]
DEFAULT_BRACKETS = SQUARE_BRACKETS + CURLY_BRACKETS + PAREN_BRACKETS

STRIP_BRACKETS_TEST_CASES = [
    (b"[[hello]]", DEFAULT_BRACKETS, b"[[hello]]"),
    ("",             DEFAULT_BRACKETS,                  ""),
    ("[]",           DEFAULT_BRACKETS,                  ""),
    ("[",            DEFAULT_BRACKETS,                  "["),
    ("[[[hello]]]",  DEFAULT_BRACKETS,                  "hello"),
    ("{([hello])}",  DEFAULT_BRACKETS,                  "hello"),
    ("({[hello]})",  DEFAULT_BRACKETS,                  "hello"),
    ("([{hello}])",  DEFAULT_BRACKETS,                  "hello"),
    ("[hello]",      DEFAULT_BRACKETS,                  "hello"),
    ("{hello}",      DEFAULT_BRACKETS,                  "hello"),
    ("(hello)",      DEFAULT_BRACKETS,                  "hello"),
    ("hello",        DEFAULT_BRACKETS,                  "hello"),
    ("{hello}",      SQUARE_BRACKETS,                   "{hello}"),
    ("{hello}",      PAREN_BRACKETS,                    "{hello}"),
    ("{hello}",      SQUARE_BRACKETS + PAREN_BRACKETS,  "{hello}"),
    ("[hello]",      CURLY_BRACKETS,                    "[hello]"),
    ("[hello]",      PAREN_BRACKETS,                    "[hello]"),
    ("[hello]",      CURLY_BRACKETS + PAREN_BRACKETS,   "[hello]"),
    ("{([",          DEFAULT_BRACKETS,                  "{(["),
    ("({[",          DEFAULT_BRACKETS,                  "({["),
    ("([{",          DEFAULT_BRACKETS,                  "([{"),
]


@pytest.mark.parametrize(
    argnames  = ["obj", "brackets", "expected"],
    argvalues = STRIP_BRACKETS_TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[1])}" for x in STRIP_BRACKETS_TEST_CASES]
)
def test_strip_brackets(obj, brackets, expected):
    assert strip_brackets(obj, brackets=brackets) == expected


STRIP_LINE_ENDING_TEST_CASES = [
    ("abc\r\n\r\ndef\r\n", "abc\r\n\r\ndef"),
    ("abc\r\n\r\ndef",     "abc\r\n\r\ndef"),
    ("abc\r\n\r\ndef\r",   "abc\r\n\r\ndef"),
    ("abc\r\n\r\ndef\n",   "abc\r\n\r\ndef"),
    ("abc\r\n\r",          "abc"),
    ("\r\n\r",             ""),
    (b"abc\r\n\r\ndef\r\n", TypeError),
    (b"abc\r\n\r\ndef",     TypeError),
    (b"abc\r\n\r\ndef\r",   TypeError),
    (b"abc\r\n\r\ndef\n",   TypeError),
    (b"abc\r\n\r",          TypeError),
    (b"\r\n\r",             TypeError),
    (None,                  AttributeError)
]


@pytest.mark.parametrize(
    argnames  = ["obj", "expected"],
    argvalues = STRIP_LINE_ENDING_TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[1])}" for x in STRIP_LINE_ENDING_TEST_CASES]
)
def test_strip_line_ending(obj, expected):
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected) as _:
            strip_line_ending(obj)
    else:
        assert strip_line_ending(obj) == expected



#!/bin/false
# -*- coding: utf-8 -*-

"""
A collection of functions for text manipulation.
"""


# =======================================================================================================================
# Main imports
# =======================================================================================================================
from typing import List, Any, Union
from ptutils.encoding import decode_json, decode_yaml
from ptutils.undefined import is_undefined
from ptutils.typing import is_string, is_sequence
from ptutils.text import strip_quotes


# =======================================================================================================================
# Utility functions: Type conversion
# =======================================================================================================================
# ------------------------------------------------------------------------------------------------------------------------
def to_string(obj: Any) -> str:
    """
    Attempt to convert object to string.
    * If `obj` is already a `str` return `obj` unmodified.
    * If `obj` is `bytes`, attempt to decode using utf-8.
    * If `obj` is any other type, return `str(obj)`.

    Parameters
    ----------
    obj : Any
        The object to convert.

    Returns
    -------
    str
        The object converted to a string.
    """
    if is_string(obj):
        return obj
    elif isinstance(obj, bytes):
        return obj.decode('utf-8')
    else:
        return str(obj)


# ------------------------------------------------------------------------------------------------------------------------
def boolify(x: Any) -> bool:
    """
    Convert object to boolean.
    * If `x` is already a `bool` return `x` unmodified.
    * If `x` is `UNDEFINED` or `None`, return False.
    * If `x` is a string, return true when `x.lower()` is one of 'yes', 'on', 'true', 'enabled', or '1'.

    Parameters
    ----------
    x : Any
        The object to inspect.

    Returns
    -------
    bool
        `True` or `False` based on best interpretation of `x`

    Raises
    ------
    TypeError
        When `x` can not be converted to a boolean.
    """
    # if the object is already a boolean, return it.
    if isinstance(x, bool):
        return x

    # if the object is None, return False
    if x is None:
        return False

    # if the object is undefined, return False
    if is_undefined(x):
        return False

    # if the object is a string, return True if it is any of the true boolean words
    if isinstance(x, str):
        x_ = x.lower()
        if x_ in ['yes', 'on', 'true', 'enabled', '1']:
            return True
        if x_ in ['no', 'off', 'false', 'disabled', '0']:
            return False

    raise TypeError("Can't convert '%s'  to boolean" % type(x).__name__)


# ------------------------------------------------------------------------------------------------------------------------
def listify(obj: Any, strict: bool = False) -> List[Any]:
    """
    Convert object to a list.
    If the object is a scalar, including astrings, return t a list with that scalar
    as the only element.
    If the object is already an sequence type, return the object unmodified unless
    `strict` is `True`, in which case convert the sequence to a `list`.

    Parameters
    ----------
    obj : Any
        The object to convert to a list
    strict : bool, optional
        When False and `obj` is a `tuple` or a `set`, return the object. When True
        and `obj` is a `tuple` or a `set`, return a list with the same contents. By
        default False

    Returns
    -------
    List[Any]
        A list of objects.
    """
    # check if the object is a sequence
    if is_sequence(obj):

        # if the sequence is not strictly a list, convert to list
        if strict and not isinstance(obj, list):
            return list(obj)

        # return the object
        return obj

    # wrap the object in a list.
    return [obj]


# ------------------------------------------------------------------------------------------------------------------------
def as_numeric(text: Any) -> Union[int, float]:
    """
    Try parsing a string as a numberic value or raise ValueError.

    Parameters
    ----------
    text : Any
        A string.

    Returns
    -------
    Union[int, float]
        The numeric value.

    Raises
    ------
    ValueError
        When `text` can not be interpreted as either an integer or floating point number.
    """
    # try parsing as float
    try:
        return float(text)
    except ValueError:
        pass

    # try parsing as int
    try:
        return int(text)
    except:  # noqa: E722
        raise ValueError("Can not parse text as an integer or float.")


# ------------------------------------------------------------------------------------------------------------------------
def as_encoded(text: Any) -> Any:
    """
    Try parsing a string as an encoded JSON or YAML string.

    Parameters
    ----------
    text : Any
        A string.

    Returns
    -------
    Any
        The decoded value.

    Raises
    ------
    ValueError
        When `text` can not be interpreted as either a JSON- or YAML-encoded string.
    """
    # try parsing as JSON
    try:
        return decode_json(text)
    except:  # noqa: E722
        pass

    # try parsing as YAML
    try:
        return decode_yaml(text)
    except:  # noqa: E722
        raise ValueError("Can not parse text as either JSON or YAML.")


# ------------------------------------------------------------------------------------------------------------------------
def as_literal(text: Any) -> Any:
    """
    Try parsing a string as bool or None literal.

    Parameters
    ----------
    text : Any
        A string.

    Returns
    -------
    Any
        The interpreted value; either a boolean or None.

    Raises
    ------
    ValueError
        When `text` can not be interpreted as either a boolean or None.
    """
    # Strip any quotes
    text2 = strip_quotes(text).lower()

    # test against falsw words
    if text2 in ['no', 'off', 'disabled', 'false']:
        return False

    # test against true words
    if text2 in ['yes', 'on', 'enabled', 'true']:
        return True

    # Check for null/none
    if text2 in ['null', 'none']:
        return None

    raise ValueError("Can not parse text as either boolean or none.")


# ------------------------------------------------------------------------------------------------------------------------
def canonize(text: Any) -> Any:
    """
    Attempt to parse a string into the most canonical pythonic form. For example, this function maps the
    strings 'yes', 'no', etc. to appropriate boolean. If a string is valid JSON or YAML(when supported), or
    a known boolean word, or integer or float, this function will convert the text to that type. If `text`
    is not a string, it is returned unmodified.

    Parameters
    ----------
    text : Any
        The string to process.

    Returns
    -------
    Any
        The most appropriate object based on the content of `text`.
    """

    # interpret bytes as utf-8
    if isinstance(text, bytes):
        return canonize(text.decode('utf-8'))

    # if not a string, return the object unmodified
    if not isinstance(text, str):
        return text

    # try parsing as numeric
    try:
        return as_numeric(text)
    except ValueError:
        pass

    # try parsing as boolean or None literal
    try:
        return as_literal(text)
    except:  # noqa: E722
        pass

    # try parsing as JSON or YAML
    try:
        return as_encoded(text)
    except:  # noqa: E722
        pass

    # give up and return the object unmodified.
    return text



#!/bin/false
# -*- coding: utf-8 -*-

"""This module provides unit tests for `ptutils.globbing`."""

import os
import pytest
from ptutils.globbing import scan_folder, get_subdirs, get_subfiles
from ._fs_fixture import GLOBBING_QUERIES, sample_file_structure, path_parts  # noqa: F401


@pytest.mark.parametrize(
    ["path", "pattern", "recursive", "filterfunc", "expected"],
    GLOBBING_QUERIES,
    ids = [f"GLOBBING_QUERIES[{i}]" for i in range(len(GLOBBING_QUERIES))]
)
def test_scan_folder(sample_file_structure, path, pattern, recursive, filterfunc, expected):  # noqa: F811
    folder = os.path.join(sample_file_structure, path)
    items = list(scan_folder(
        folder     = folder,
        pattern    = pattern,
        recursive  = recursive,
        filterfunc = filterfunc
    ))
    items_rel    = set([os.path.relpath(item, sample_file_structure) for item in items])
    expected_rel = set([os.path.join(*path_parts(path)) for path in expected])
    assert bool(items_rel)
    assert bool(expected_rel)
    assert items_rel == expected_rel


@pytest.mark.parametrize(
    ["path", "pattern", "recursive", "filterfunc", "expected"],
    GLOBBING_QUERIES,
    ids = [f"GLOBBING_QUERIES[{i}]" for i in range(len(GLOBBING_QUERIES))]
)
def test_get_subdirs(sample_file_structure, path, pattern, recursive, filterfunc, expected):  # noqa: F811
    if filterfunc is os.path.isdir:
        folder = os.path.join(sample_file_structure, path)
        items = list(get_subdirs(
            folder     = folder,
            pattern    = pattern,
            recursive  = recursive
        ))
        items_rel    = set([os.path.relpath(item, sample_file_structure) for item in items])
        expected_rel = set([os.path.join(*path_parts(path)) for path in expected])
        assert bool(items_rel)
        assert bool(expected_rel)
        assert items_rel == expected_rel


@pytest.mark.parametrize(
    ["path", "pattern", "recursive", "filterfunc", "expected"],
    GLOBBING_QUERIES,
    ids = [f"GLOBBING_QUERIES[{i}]" for i in range(len(GLOBBING_QUERIES))]
)
def test_get_subfiles(sample_file_structure, path, pattern, recursive, filterfunc, expected):  # noqa: F811
    if filterfunc is os.path.isfile:
        folder = os.path.join(sample_file_structure, path)
        items = list(get_subfiles(
            folder     = folder,
            pattern    = pattern,
            recursive  = recursive
        ))
        items_rel    = set([os.path.relpath(item, sample_file_structure) for item in items])
        expected_rel = set([os.path.join(*path_parts(path)) for path in expected])
        assert bool(items_rel)
        assert bool(expected_rel)
        assert items_rel == expected_rel



#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides unit tests for `ptutils.typing`.
"""

import pytest
from ptutils.typing import is_iterable, is_sequence, is_string, is_mapping


TEST_CASES = [
    ("a",               False, True,  True,  False),
    (b"a",              False, False, True,  True ),
    (None,              False, False, False, False),
    ([None],            False, False, True,  True),
    (set([None]),       False, False, True,  True),
    (tuple([None]),     False, False, True,  True),
    (iter([None]),      False, False, True,  True),
    (dict(a='b'),       True,  False, True,  False),
]


@pytest.mark.parametrize(
    argnames  = ["obj", "obj_is_mapping", "obj_is_string", "obj_is_iterable", "obj_is_sequence"],
    argvalues = TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[1])}" for x in TEST_CASES]
)
def test_is_mapping(obj, obj_is_mapping, obj_is_string, obj_is_iterable, obj_is_sequence):
    expected = obj_is_mapping
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected) as _:
            is_mapping(obj)
    else:
        assert is_mapping(obj) == expected


@pytest.mark.parametrize(
    argnames  = ["obj", "obj_is_mapping", "obj_is_string", "obj_is_iterable", "obj_is_sequence"],
    argvalues = TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[2])}" for x in TEST_CASES]
)
def test_is_string(obj, obj_is_mapping, obj_is_string, obj_is_iterable, obj_is_sequence):
    expected = obj_is_string
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected) as _:
            is_string(obj)
    else:
        assert is_string(obj) == expected


@pytest.mark.parametrize(
    argnames  = ["obj", "obj_is_mapping", "obj_is_string", "obj_is_iterable", "obj_is_sequence"],
    argvalues = TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[3])}" for x in TEST_CASES]
)
def test_is_iterable(obj, obj_is_mapping, obj_is_string, obj_is_iterable, obj_is_sequence):
    expected = obj_is_iterable
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected) as _:
            is_iterable(obj)
    else:
        assert is_iterable(obj) == expected


@pytest.mark.parametrize(
    argnames  = ["obj", "obj_is_mapping", "obj_is_string", "obj_is_iterable", "obj_is_sequence"],
    argvalues = TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[3])}" for x in TEST_CASES]
)
def test_is_sequence(obj, obj_is_mapping, obj_is_string, obj_is_iterable, obj_is_sequence):
    expected = obj_is_sequence
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected) as _:
            is_sequence(obj)
    else:
        assert is_sequence(obj) == expected



#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides unit tests for `ptutils.time`.
"""

import datetime
from ptutils.time import now, dt


def test_now():
    assert isinstance(now(), datetime.datetime)


def test_dt():
    t1 = now()
    t2 = now()
    d = dt(t1, t2)
    assert isinstance(d, datetime.timedelta)
    assert d.total_seconds() > 0

    d2 = dt(t1)
    assert isinstance(d2, datetime.timedelta)
    assert d2.total_seconds() > 0



#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides unit tests for `ptutils.lazy`.
"""

from ptutils.lazy import lazy_property


def test_lazy_property():
    marker = []

    def _inner():
        marker.append(0)
        return True

    assert len(marker) == 0

    class Klass:
        @lazy_property
        def prop(self):
            return _inner()

    c = Klass()
    assert isinstance(c, Klass)
    assert len(marker) == 0
    p = c.prop
    assert len(marker) == 1
    assert p is True
    assert marker == [0]
    p = c.prop
    assert len(marker) == 1
    assert p is True
    assert marker == [0]



#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides unit tests for `ptutils.io`.
"""


import os
import tempfile
import pytest
import shutil
from ptutils.encoding import (
    HAVE_YAML,
    encode_json,
    decode_json,
    encode_yaml,
    decode_yaml
)
from ptutils.globbing import scan_folder
from ptutils.io import (
    get_path_part,
    get_hash_tmp_dir,
    get_filetype,
    makedirs,
    read,
    read_lines,
    read_all,
    write_all,
    HAVE_TAR,
    require_tar,
    extract_archive,
    compress_archive,
    load_txt,
    save_txt,
    load_json,
    save_json,
    load_yaml,
    save_yaml,
    get_encoder,
    get_decoder,
    get_loader,
    get_saver,
    load,
    save,
)
from ._fs_fixture import (  # noqa: F401
    localize_path,
    sample_file_structure,
    FILE_STRUCTURE
)


FILE_TYPE_TEST_CASES = [
    ("/etc/myapp/config.json", "json"),
    ("/etc/myapp/config.jsn", "json"),
    ("/etc/myapp/config.yaml", "yaml"),
    ("/etc/myapp/config.yml", "yaml"),
    ("/etc/myapp.json/config.yml", "yaml"),
    ("/etc/myapp.yaml/config.jsn", "json"),
    ("/etc/myapp/data.zip", "archive-zip"),
    ("/etc/myapp/data.tar", "archive-tar"),
    ("/etc/myapp/data.bz2", "archive-bzip2"),
    ("/etc/myapp/data.gz", "archive-gzip"),
    ("/etc/myapp/data.tar.bz2", "archive-bzip2"),
    ("/etc/myapp/data.tar.gz", "archive-gzip"),
    ("myapp/config.json", "json"),
    ("myapp/config.jsn", "json"),
    ("myapp/config.yaml", "yaml"),
    ("myapp/config.yml", "yaml"),
    ("myapp.json/config.yml", "yaml"),
    ("myapp.yaml/config.jsn", "json"),
    ("myapp/data.zip", "archive-zip"),
    ("myapp/data.tar", "archive-tar"),
    ("myapp/data.bz2", "archive-bzip2"),
    ("myapp/data.gz", "archive-gzip"),
    ("myapp/data.tar.bz2", "archive-bzip2"),
    ("myapp/data.tar.gz", "archive-gzip"),
    ("config.json", "json"),
    ("config.jsn", "json"),
    ("config.yaml", "yaml"),
    ("config.yml", "yaml"),
    ("config.yml", "yaml"),
    ("config.jsn", "json"),
    ("data.zip", "archive-zip"),
    ("data.tar", "archive-tar"),
    ("data.bz2", "archive-bzip2"),
    ("data.gz", "archive-gzip"),
    ("data.tar.bz2", "archive-bzip2"),
    ("data.tar.gz", "archive-gzip"),
]


@pytest.mark.parametrize(
    argnames  = ["path", "expected"],
    argvalues = FILE_TYPE_TEST_CASES,
    ids       = [f"{x[0]},type={x[1]}" for x in FILE_TYPE_TEST_CASES]
)
def test_get_file_type(path, expected):
    filename = localize_path(path)
    assert get_filetype(filename) == expected


PATH_PART_TEST_CASES = [
    (
        "/a/b/c.tar.gz",
        {
            "fullpath":  "/a/b/c.tar.gz",
            "abspath":   "/a/b/c.tar.gz",
            "name":      "c.tar.gz",
            "extension": "gz",
            "basename":  "c.tar"
        }
    ),
    (
        "/a/b/../c.tar.bz2",
        {
            "fullpath":  "/a/b/../c.tar.bz2",
            "abspath":   "/a/c.tar.bz2",
            "name":      "c.tar.bz2",
            "extension": "bz2",
            "basename":  "c.tar"
        }
    ),
    (
        "c.zip",
        {
            "fullpath":  "c.zip",
            "abspath":   os.path.abspath(os.path.join(os.curdir, "c.zip")),
            "name":      "c.zip",
            "extension": "zip",
            "basename":  "c"
        }
    ),
    (
        "/README",
        {
            "fullpath":  "/README",
            "abspath":   os.path.abspath("/README"),
            "name":      "README",
            "extension": '',
            "basename":  "README"
        }
    ),
]
PATH_PART_LABELS = [
    "fullpath",
    "abspath",
    "name",
    "extension",
    "basename",
]


@pytest.mark.parametrize(
    argnames = "part",
    argvalues = PATH_PART_LABELS,
    ids       = [f"PART={label}" for label in PATH_PART_LABELS]
)
@pytest.mark.parametrize(
    argnames = ["path", "expectations"],
    argvalues = PATH_PART_TEST_CASES,
    ids       = [ x[0] for x in PATH_PART_TEST_CASES ]
)
def test_get_path_part(path, expectations, part):
    filename = localize_path(path)
    result   = localize_path(expectations[part])
    assert get_path_part(filename=filename, part=part) == result


def test_get_path_part_raises_for_unknown_part():
    with pytest.raises(ValueError) as _:
        get_path_part(filename="/path/to/a/file.ext", part="unknown")


# -----------------------------------------------------------------------------------------------------------------------
# Globals
# -----------------------------------------------------------------------------------------------------------------------
__FILETYPES__ = {
    'archive-zip': ['zip'],
    'archive-tar': ['tar'],
    'archive-bzip2': ['bz2'],
    'archive-gzip':  ['gz'],
    'yaml': ['yml', 'yaml'],
    'json': ['jsn', 'json']
}


def test_makedirs_creates_all(sample_file_structure):  # noqa: F811
    parts = ["apples", "lesser-known", "from-mexico", "juarez"]
    folder = os.path.join(localize_path('/'.join(parts)))
    makedirs(folder)
    for i in range(1, len(parts)):
        assert os.path.isdir(
            os.path.join(localize_path('/'.join(parts[:i])))
        )


def test_makedirs_fails_when_a_parent_is_a_file(sample_file_structure):  # noqa: F811
    parts = ["apples", "granny-smith", "lesser-known", "from-mexico", "juarez"]
    folder = os.path.join(sample_file_structure, localize_path('/'.join(parts)))
    filename = os.path.join(sample_file_structure, "apples/granny-smith")
    assert os.path.isfile(filename)
    with pytest.raises(OSError) as _:
        makedirs(folder)
    assert os.path.isdir(os.path.dirname(filename))
    assert os.path.isfile(filename)


def test_read(sample_file_structure):  # noqa: F811
    folder = os.path.join(sample_file_structure, 'boring_things')
    f1     = os.path.join(folder, 'thing.json')
    assert read(f1) == FILE_STRUCTURE['boring_things']['thing.json']
    f2     = os.path.join(folder, 'text_thing')
    assert read(f2) == FILE_STRUCTURE['boring_things']['text_thing']
    assert read_all(f1) == FILE_STRUCTURE['boring_things']['thing.json']
    assert read_all(f2) == FILE_STRUCTURE['boring_things']['text_thing']
    assert load_txt(f1) == FILE_STRUCTURE['boring_things']['thing.json']
    assert load_txt(f2) == FILE_STRUCTURE['boring_things']['text_thing']


def test_write(sample_file_structure):  # noqa: F811
    folder = os.path.join(sample_file_structure, 'boring_things', 'tbd_things')
    f1     = os.path.join(folder, 'tbd.txt')
    write_all(f1, FILE_STRUCTURE['boring_things']['text_thing'])
    assert read(f1) == FILE_STRUCTURE['boring_things']['text_thing']


def test_txt(sample_file_structure):  # noqa: F811
    folder = os.path.join(sample_file_structure, 'boring_things', 'tbd_things')
    f1     = os.path.join(folder, 'tbd.txt')
    save_txt(f1, FILE_STRUCTURE['boring_things']['text_thing'])
    assert load_txt(f1) == FILE_STRUCTURE['boring_things']['text_thing']


def test_get_encoder():
    assert get_encoder("x.json") is encode_json
    assert get_encoder("x.jsn") is encode_json
    assert get_encoder("x.yaml") is encode_yaml
    assert get_encoder("x.yml") is encode_yaml
    with pytest.raises(Exception) as _:
        get_encoder("x.txt")
    with pytest.raises(Exception) as _:
        get_encoder("x")


def test_get_decoder():
    assert get_decoder("x.json") is decode_json
    assert get_decoder("x.jsn") is decode_json
    assert get_decoder("x.yaml") is decode_yaml
    assert get_decoder("x.yml") is decode_yaml
    with pytest.raises(Exception) as _:
        get_decoder("x.txt")
    with pytest.raises(Exception) as _:
        get_decoder("x")


def test_get_loader():
    assert get_loader("x.json") is load_json
    assert get_loader("x.jsn") is load_json
    assert get_loader("x.yaml") is load_yaml
    assert get_loader("x.yml") is load_yaml
    with pytest.raises(Exception) as _:
        get_loader("x.txt")

    assert get_loader("x", default=load_txt) is load_txt


def test_get_saver():
    assert get_saver("x.json") is save_json
    assert get_saver("x.jsn") is save_json
    assert get_saver("x.yaml") is save_yaml
    assert get_saver("x.yml") is save_yaml
    with pytest.raises(Exception) as _:
        get_saver("x.txt")

    assert get_saver("x", default=save_txt) is save_txt


STRUCTURED_DATA = {
    "a_bool": True,
    "An_int": 123,
    "An_float": 123.456,
    "a string": "hello world",
    "A list of things": [
        True,
        None,
        "hello",
        [],
        123,
        456.789
    ],
    "a dict": {
        "a": 123,
        "b": True,
        "c": "hi",
        "d": dict(),
        "e": list()
    }
}


def test_round_trip_json_file(sample_file_structure):  # noqa: F811
    f1 = os.path.join(sample_file_structure, 'something.json')
    save_json(f1, STRUCTURED_DATA)
    assert load_json(f1) == STRUCTURED_DATA


def test_round_trip_yaml_file(sample_file_structure):  # noqa: F811
    if HAVE_YAML:
        f1 = os.path.join(sample_file_structure, 'something.yaml')
        save_yaml(f1, STRUCTURED_DATA)
        assert load_yaml(f1) == STRUCTURED_DATA
    else:
        pytest.skip("No YAML support")


def test_round_trip_save_load(sample_file_structure):  # noqa: F811
    f1 = os.path.join(sample_file_structure, 'something.json')
    save(f1, STRUCTURED_DATA)
    assert load(f1) == STRUCTURED_DATA

    if HAVE_YAML:
        f1 = os.path.join(sample_file_structure, 'something.yaml')
        save(f1, STRUCTURED_DATA)
        assert load(f1) == STRUCTURED_DATA
    else:
        pytest.skip("No YAML support")


def test_get_hash_tmp_dir():
    for case in [
        "",
        "a/b/c",
        "/a/b"
    ]:
        d1 = get_hash_tmp_dir(case)
        d2 = os.path.dirname(d1)
        assert d2 == tempfile.gettempdir()
        assert isinstance(d1, str)
        assert bool(d1)
        assert not os.path.isdir(d1)
        assert os.path.isdir(d2)


def test_read_lines(sample_file_structure):  # noqa: F811
    content_lines = [
        "rootfs / lxfs rw,noatime 0 0",
        "none /dev tmpfs rw,noatime,mode=755 0 0",
        "sysfs /sys sysfs rw,nosuid,nodev,noexec,noatime 0 0",
        "proc /proc proc rw,nosuid,nodev,noexec,noatime 0 0",
        "devpts /dev/pts devpts rw,nosuid,noexec,noatime,gid=5,mode=620 0 0",
        "none /run tmpfs rw,nosuid,noexec,noatime,mode=755 0 0",
        "none /run/lock tmpfs rw,nosuid,nodev,noexec,noatime 0 0",
        "none /run/shm tmpfs rw,nosuid,nodev,noatime 0 0",
        "none /run/user tmpfs rw,nosuid,nodev,noexec,noatime,mode=755 0 0",
        "binfmt_misc /proc/sys/fs/binfmt_misc binfmt_misc rw,relatime 0 0",
        "tmpfs /sys/fs/cgroup tmpfs rw,nosuid,nodev,noexec,relatime,mode=755 0 0",
        "cgroup /sys/fs/cgroup/devices cgroup rw,nosuid,nodev,noexec,relatime,devices 0 0"
    ]
    content = os.linesep.join(content_lines)
    filename = os.path.join(sample_file_structure, 'boring_things', 'tbd_things', 'a-text-file.txt')
    write_all(filename, content)
    assert read(filename) == content
    assert read_lines(filename) == content_lines


def test_require_tar():
    if HAVE_TAR:
        require_tar()
    else:
        with pytest.raises(Exception) as _:
            require_tar()


def test_archive_round_trip(sample_file_structure):  # noqa: F811
    with pytest.raises(NotImplementedError) as _:
        extract_archive(archive=localize_path("/file.zip"), folder=localize_path("/tmp"))
    with pytest.raises(NotImplementedError) as _:
        extract_archive(archive=localize_path("/file.7z"), folder=localize_path("/tmp"))
    if not HAVE_TAR:
        with pytest.raises(Exception) as _:
            extract_archive(archive=localize_path("/file.tar.gz"), folder=localize_path("/tmp"))
    else:
        source_folder   = sample_file_structure
        source_contents = [ os.path.relpath(x, source_folder) for x in list(scan_folder(source_folder))]

        archive         = os.path.join(tempfile.gettempdir(), 'something.tar.gz')
        dest_folder     = os.path.join(tempfile.gettempdir(), 'something.tar.gz.contents')
        try:
            compress_archive(archive=archive, folder=source_folder)
            assert os.path.isfile( archive )
            extract_archive(archive=archive, folder=dest_folder)
            assert os.path.isdir( dest_folder )
            dest_contents = [ os.path.relpath(x, dest_folder) for x in list(scan_folder(dest_folder))]
            assert dest_contents == source_contents
        finally:
            try:
                os.unlink(archive)
            except:  # noqa: E722
                pass
            try:
                shutil.rmtree(dest_contents)
            except:  # noqa: E722
                pass



#!/bin/false
# -*- coding: utf-8 -*-

""" Tolls for JSON and YAML encoding/decoding. """


# ------------------------------------------------------------------------------------------------------------------------
# Main imports
# ------------------------------------------------------------------------------------------------------------------------
import datetime
import json
import os
from types import ModuleType
from typing import Any, Union, IO

# ------------------------------------------------------------------------------------------------------------------------
# Import checks
# ------------------------------------------------------------------------------------------------------------------------
try:
    import yaml
    HAVE_YAML = True
except ImportError:  # pragma: no cover
    HAVE_YAML = False

# ------------------------------------------------------------------------------------------------------------------------
# Debug functions
# ------------------------------------------------------------------------------------------------------------------------
"""
Global flag for whether or not to log encoding error debugging info when encoding fails.
This is initialized with the content of environment variable 'PTUTILS_ENABLE_ENCODER_DEBUG'.
"""
PTUTILS_ENABLE_ENCODER_DEBUG = (
    os.environ.get('PTUTILS_ENABLE_ENCODER_DEBUG', '').lower()
    in
    ['yes', 'true', 'on', 'enabled']
)


# ------------------------------------------------------------------------------------------------------------------------
def debug_encoder_failure(obj: Any) -> None:

    if PTUTILS_ENABLE_ENCODER_DEBUG:  # pragma: no cover
        # dont create a logger unless we need it...
        from ptutils.logging import getLogger
        logger = getLogger(__name__)

        # dump some useful information about the failure
        try:
            logger.debug('#' * 80)
            logger.debug('# JSON Encoding error. Unable to encode object as JSON.')
            logger.debug("# Object details:")
            logger.debug("#   class name:     %s" % type(obj).__name__)
            logger.debug("#   Representation: %s" % repr(obj))
            logger.debug("#   String:         %s" % str(obj))
            logger.debug('#' * 80)
        except:  # noqa E722 # pylint: disable=broad-except
            logger.exception("Error debugging object JSON serialization failure.")


# ------------------------------------------------------------------------------------------------------------------------
# Utility functions
# ------------------------------------------------------------------------------------------------------------------------
def json_serial(obj: Any) -> Any:
    """
    JSON serialization helper for objects not serializable by default.

    Parameters
    ----------
    obj : Any
        The object to be serialized

    Returns
    -------
    Any
        An object which is serializable by the `json` python standard library.

    Raises
    ------
    TypeError
        If the object could not be converted to a serializable type.

    Notes
    -----
    This function provides serialization for `datetime.datetime`,
    `datetime.timedelta`, and `datetime.date`.
    For other types, a TypeError will be raised, and some debugging information
    about the encoding failure will be logged using `logger.debug`.

    """
    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    elif isinstance(obj, datetime.timedelta):
        return obj.total_seconds()

    debug_encoder_failure(obj)

    raise TypeError(f"Type {type(obj).__name__} not serializable.")


# ------------------------------------------------------------------------------------------------------------------------
def encode_json(obj: Any) -> str:
    """
    Encode the object as a JSON string

    Parameters
    ----------
    obj : Any
        The object to encode

    Returns
    -------
    str
        A JSON string representation of the object.
    """
    return json.dumps(obj, sort_keys=True, indent=4, separators=[', ', ': '], default=json_serial)


# ------------------------------------------------------------------------------------------------------------------------
def pretty_json(obj: Any) -> str:
    """
    Encode the object as a JSON string in human-friendly format.

    Parameters
    ----------
    obj : Any
        The object to encode

    Returns
    -------
    str
        A JSON string representation of the object.
    """
    return encode_json(obj)


# ------------------------------------------------------------------------------------------------------------------------
def decode_json(text: str) -> Any:
    """
    Encode a JSON string into a python object.

    Parameters
    ----------
    text : str
        The JSON text to parse.

    Returns
    -------
    Any
        The decoded object.
    """
    return json.loads(text)


# ------------------------------------------------------------------------------------------------------------------------
def require_yaml() -> ModuleType:
    """
    Ensure that YAML encoder is installed or else raise an exception.

    Returns
    -------
    ModuleType
        The yaml module itself.

    Raises
    ------
    Exception
        When YAML encoder can not be imported.
    """
    if not HAVE_YAML:  # pragma: no cover
        raise Exception("YAML encoder is not available. Try 'pip install pyyaml'")

    return yaml


# ------------------------------------------------------------------------------------------------------------------------
def decode_yaml(text_or_stream: Union[str, IO]) -> Any:
    """
    Decode a YAML string into a python object.

    Parameters
    ----------
    text_or_stream : Union[str, IO]
        Either a string literal containing YAML formatted markup, or an object
        providing the stream protocol whose contents will be read and then parsed.

    Returns
    -------
    Any
        The python object described by the yaml text.

    Raises
    ------
    Exception
        When YAML encoder can not be imported.
    """
    require_yaml()
    return yaml.safe_load(text_or_stream)


# ------------------------------------------------------------------------------------------------------------------------
def encode_yaml(obj: Any) -> str:
    """
    Serialize an object to a YAML-formatted string.

    Parameters
    ----------
    obj : Any
        The object to be serialized.

    Returns
    -------
    str
        The YAML-formatted string representation of `obj`

    Raises
    ------
    Exception
        When YAML encoder can not be imported.
    """
    require_yaml()
    return yaml.safe_dump(obj, explicit_start=True, default_flow_style=False)


# ------------------------------------------------------------------------------------------------------------------------
def pretty_yaml(obj: Any) -> str:
    """
    Serialize an object to a human-friendly, YAML-formatted string.

    Parameters
    ----------
    obj : Any
        The object to be serialized.

    Returns
    -------
    str
        The YAML-formatted string representation of `obj`

    Raises
    ------
    Exception
        When YAML encoder can not be imported.
    """
    return encode_yaml(obj)



#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides functions and classes for reflection and introspection.
"""

# -----------------------------------------------------------------------------------------------------------------------
# Main imports
# -----------------------------------------------------------------------------------------------------------------------
import inspect
import re
import sys
from typing import Optional, Callable, Dict


# -----------------------------------------------------------------------------------------------------------------------
# Utility functions
# -----------------------------------------------------------------------------------------------------------------------
def get_function(name: str, module: str = '__main__') -> Optional[Callable]:
    """ Get the function named `name` from module names `module`.

    Parameters
    ----------
    name: str
        The name of the function.
    module: str, optional
        The name of the module in which to find the function `name`.

    Returns
    -------
    func: Callable, optional
        The function object if present, None otherwise.
    """
    for name_, function in inspect.getmembers(sys.modules[module]):
        if inspect.isfunction(function):
            if name_ == name:
                return function
    return None


# -----------------------------------------------------------------------------------------------------------------------
def get_functions(regex: str, expander: Optional[str] = None, module: str = '__main__') -> Dict[str, Callable]:
    """ Find all functions in a module with names matching the specified regular expression.
    Additionally, the dictionary keys may be modified by supplying an expander.

    Parameters
    ----------
    regex: str
        The regular expression with which to match function names.
    expander: str, optional
        A regular expression expansion string. When supplied, the function's key in the returned dictionary
        will be computed from the match object's expand() method.
    module: str, optional
        The name of the module in which to find the functions matching `regex`.

    Returns
    -------
    func: Callable, optional
        The function object if present, None otherwise.
    """
    rv = {}
    for name, function in inspect.getmembers(sys.modules[module]):
        if inspect.isfunction(function):
            m = re.match(regex, name)
            if m:
                if expander:
                    rv[m.expand(expander)] = function
                else:
                    rv[name] = function
    return rv
